@@ -20,7 +20,7 @@ Contact: yuvraaj@gmail.com
 */
 
 #include "GVApi.h"
-#include "GvXMLParser.h"
+#include "HtmlFieldParser.h"
 #include "MyXmlErrorHandler.h"
 
 #define DEBUG_ONLY 0
@@ -412,6 +412,15 @@ GVApi::cancel(AsyncTaskToken *token)
     tracker->abort ();
 }//GVApi::cancel
 
+/*
+ * GET http://google.com/voice
+ * Should result in a bunch of redirects (to FQDN, then to mobile https)
+ * eventually landing on
+ * https://accounts.google.com/ServiceLogin?service=grandcentral
+ *      &continue=https://www.google.com/voice/m?initialauth
+ *      &followup=https://www.google.com/voice/m?initialauth
+ * ... which is handled by onLogin1.
+ */
 bool
 GVApi::login(AsyncTaskToken *token)
 {
@@ -449,6 +458,10 @@ GVApi::login(AsyncTaskToken *token)
     return rv;
 }//GVApi::login
 
+/*
+ * Parse out the login form out of the Service Login page.
+ * "Fill up" the form and post it with postLogin.
+ */
 void
 GVApi::onLogin1(bool success, const QByteArray &response, QNetworkReply *reply,
                 void *ctx)
@@ -506,8 +519,9 @@ GVApi::onLogin1(bool success, const QByteArray &response, QNetworkReply *reply,
             break;
         }
         if (!nextAction.startsWith ("http")) {
-            Q_DEBUG(QString("nextAction %1").arg (nextAction));
-            nextAction = GOOGLE_ACCOUNTS "/" + nextAction;
+            QUrl nextUrl = reply->url().resolved (nextAction);
+            nextAction = nextUrl.toString ();
+            Q_DEBUG(nextAction);
         }
 
         Q_DEBUG(QString("Starting service login by posting login to %1")
@@ -546,10 +560,10 @@ GVApi::parseHiddenLoginFields(const QString &strResponse, QVariantMap &ret)
   <input type="hidden" name="continue" id="continue"
            value="https://www.google.com/voice/m" />
 */
-    QRegExp rx1("\\<input\\s*type\\s*=\\s*\"hidden\"(.*)\\>");
+    QRegExp rx1("\\<input(.*)\\>");
     rx1.setMinimal (true);
     if (!strResponse.contains (rx1)) {
-        Q_WARN("Invalid login page!");
+        Q_WARN("Invalid login page: No input fields");
         return false;
     }
 
@@ -559,33 +573,53 @@ GVApi::parseHiddenLoginFields(const QString &strResponse, QVariantMap &ret)
         QString fullMatch = rx1.cap(0);
         QString oneInstance = rx1.cap(1);
         QString name, value;
-        QRegExp rx2("\"(.*)\""), rx3("'(.*)'");
-        rx2.setMinimal (true);
-        rx3.setMinimal (true);
+        QVariantMap attrs;
 
-        int pos1 = oneInstance.indexOf ("value");
-        if (pos1 == -1) {
+        if (!fullMatch.endsWith("/>")) {
+            fullMatch = fullMatch.mid(0, fullMatch.length()-1) + "/>";
+        }
+
+        QXmlInputSource inputSource;
+        QXmlSimpleReader simpleReader;
+        inputSource.setData (fullMatch);
+        HtmlFieldParser xmlHandler;
+        xmlHandler.setEmitLog (emitLog);
+
+        simpleReader.setContentHandler (&xmlHandler);
+        simpleReader.setErrorHandler (&xmlHandler);
+
+        if (!simpleReader.parse (&inputSource, false)) {
+            Q_WARN("Failed to parse input field.");
             goto gonext;
         }
 
-        name  = oneInstance.left (pos1);
-        value = oneInstance.mid (pos1);
-
-        if (rx2.indexIn (name) == -1) {
+        if (!xmlHandler.elems.contains ("input") || 
+            !xmlHandler.attrMap.contains("input")) {
+            Q_WARN("Failed to parse input field.");
             goto gonext;
         }
-        name = rx2.cap (1);
 
-        if (rx2.indexIn (value) == -1) {
-            if (rx3.indexIn (value) == -1) {
-                goto gonext;
-            } else {
-                value = rx3.cap (1);
+        attrs = xmlHandler.attrMap["input"];
+        if (!attrs.contains ("name") ||
+            !attrs.contains ("value") ||
+            !attrs.contains ("type")) {
+            Q_WARN(QString("Input field \"%1\" doesn't have name/type/value")
+                    .arg(oneInstance));
+            goto gonext;
+        }
+        if (attrs["type"].toString() != "hidden") {
+#if DEBUG_ONLY
+            if (emitLog) {
+                Q_DEBUG(QString("Input field \"%1\" is not hidden")
+                    .arg(oneInstance));
             }
-        } else {
-            value = rx2.cap (1);
+#endif
+            goto gonext;
         }
 
+        name  = attrs["name"].toString ();
+        value = attrs["value"].toString ();
+
 #if DEBUG_ONLY
         if (ret.contains (name) &&
            (ret[name].toString() != value)) {
@@ -593,6 +627,7 @@ GVApi::parseHiddenLoginFields(const QString &strResponse, QVariantMap &ret)
                     .arg (name, ret[name].toString(), value));
         }
 #endif
+
         ret[name] = value;
 
 gonext:
@@ -664,6 +699,96 @@ GVApi::postLogin(QUrl url, AsyncTaskToken *token)
     return found;
 }//GVApi::postLogin
 
+bool
+GVApi::parseAlternateLogins(const QString &form, AsyncTaskToken *task)
+{
+/* To match:
+  <input type="radio" name="retry" id="SMS_..."   value="SMS_..." />
+  <input type="radio" name="retry" id="VOICE_..." value="VOICE_" />
+*/
+    QRegExp rx1("\\<input(.*)\\>");
+    rx1.setMinimal (true);
+    if (!form.contains (rx1)) {
+        Q_WARN("No input fields");
+        return false;
+    }
+
+    int pos = 0;
+    while ((pos = rx1.indexIn (form, pos)) != -1) {
+        QString fullMatch = rx1.cap(0);
+        QString oneInstance = rx1.cap(1);
+        QString name, value;
+        QVariantMap attrs;
+
+        if (!fullMatch.endsWith("/>")) {
+            fullMatch = fullMatch.mid(0, fullMatch.length()-1) + "/>";
+        }
+
+        QXmlInputSource inputSource;
+        QXmlSimpleReader simpleReader;
+        inputSource.setData (fullMatch);
+        HtmlFieldParser xmlHandler;
+        xmlHandler.setEmitLog (emitLog);
+
+        simpleReader.setContentHandler (&xmlHandler);
+        simpleReader.setErrorHandler (&xmlHandler);
+
+        if (!simpleReader.parse (&inputSource, false)) {
+            Q_WARN("Failed to parse input field.");
+            goto gonext;
+        }
+
+        if (!xmlHandler.elems.contains ("input") || 
+            !xmlHandler.attrMap.contains("input")) {
+                Q_WARN("Failed to parse input field.");
+                goto gonext;
+        }
+
+        attrs = xmlHandler.attrMap["input"];
+        if (!attrs.contains ("name") ||
+            !attrs.contains ("value") ||
+            !attrs.contains ("type")) {
+                Q_WARN(QString("Input field \"%1\" doesn't have name/type/value")
+                    .arg(oneInstance));
+                goto gonext;
+        }
+        if (attrs["type"].toString() != "radio") {
+#if DEBUG_ONLY
+            if (emitLog) {
+                Q_DEBUG(QString("Input field \"%1\" is not a radio")
+                    .arg(oneInstance));
+            }
+#endif
+            goto gonext;
+        }
+
+        name  = attrs["name"].toString ();
+        value = attrs["value"].toString ();
+
+        if (name != "retry") {
+            goto gonext;
+        }
+
+        if (value.startsWith("VOICE_")) {
+            task->inParams["tfaAlternate"] = value;
+        }
+        if (value.startsWith("SMS_")) {
+            task->inParams["tfaAltSMS"] = value;
+        }
+
+gonext:
+        pos += fullMatch.indexOf (oneInstance);
+    }
+
+    // This is the only MUST return field
+    if (!task->inParams.contains("tfaAlternate")) {
+        return false;
+    }
+
+    return true;
+
+}//GVApi::parseAlternateLogins
+
 void
 GVApi::onLogin2(bool success, const QByteArray &response, QNetworkReply *reply,
                 void *ctx)
@@ -685,10 +810,6 @@ GVApi::onLogin2(bool success, const QByteArray &response, QNetworkReply *reply,
     fTemp.close ();
 #endif
 
-    if (strResponse.contains ("FinishSignIn")) {
-        Q_DEBUG("Here!!");
-    }
-
     token->errorString.clear();
     do {
         if (!success) {
@@ -759,7 +880,7 @@ GVApi::onLogin2(bool success, const QByteArray &response, QNetworkReply *reply,
             quint32 len = cap.length ();
             QRegExp rxAction("action=[\"|'](.*)[\"|']");
             rxAction.setMinimal (true);
-            if (cap.contains ("verify-form")) {
+            if (cap.contains ("gaia_secondfactorform")) {
                 if (cap.contains (rxAction)) {
                     nextAction = rxAction.cap (1);
                     break;
@@ -774,30 +895,15 @@ GVApi::onLogin2(bool success, const QByteArray &response, QNetworkReply *reply,
             success = false;
             break;
         }
+        if (!nextAction.startsWith ("http")) {
+            QUrl nextUrl = replyUrl.resolved (nextAction);
+            nextAction = nextUrl.toString ();
+            Q_DEBUG(nextAction);
+        }
 
-        do {
-            pos = strResponse.indexOf ("alternative-delivery");
-            if (-1 == pos) {
-                Q_WARN("Failed to get alternate delivery");
-                break;
-            }
-            pos = strResponse.lastIndexOf ('<', pos);
-            if (-1 == pos) {
-                Q_WARN("Failed to get alternate delivery");
-                break;
-            }
-            cap = strResponse.mid(pos, strResponse.indexOf('>', pos) - pos);
-
-            QRegExp rxHref("href\\s*=\\s*\"(.*)\"");
-            rxHref.setMinimal (true);
-            pos = cap.indexOf (rxHref);
-            if (-1 == pos) {
-                Q_WARN("Failed to get alternate delivery href");
-                break;
-            }
-            cap = rxHref.cap (1);
-            token->inParams["tfaAlternate"] = cap;
-        } while(0);
+        if (!parseAlternateLogins (rxForm.cap(0), token)) {
+            Q_WARN("Failed to get alternate delivery");
+        }
 
         if (emitLog) {
             Q_DEBUG("Two factor AUTH required!");
@@ -1145,15 +1251,21 @@ GVApi::onGetPhones(bool success, const QByteArray &response, QNetworkReply *,
         QXmlInputSource inputSource;
         QXmlSimpleReader simpleReader;
         inputSource.setData (strReply);
-        GvXMLParser xmlHandler;
+        HtmlFieldParser xmlHandler;
         xmlHandler.setEmitLog (emitLog);
 
         simpleReader.setContentHandler (&xmlHandler);
         simpleReader.setErrorHandler (&xmlHandler);
         simpleReader.parse (&inputSource, false);
 
+        if (!xmlHandler.elems.contains ("json") ||
+            !xmlHandler.elems.contains ("html")) {
+            Q_WARN("Couldn't parse either the JSON or the HTML");
+            break;
+        }
+
         QString strTemp;
-        strTemp = "var o = " + xmlHandler.strJson;
+        strTemp = "var o = " + xmlHandler.elems["json"].toString();
         scriptEngine.evaluate (strTemp);
         if (scriptEngine.hasUncaughtException ()) {
             strTemp = QString ("Could not assign json to obj : %1")
@@ -1389,7 +1501,7 @@ GVApi::onGetInbox(bool success, const QByteArray &response, QNetworkReply *,
         QXmlInputSource inputSource;
         QXmlSimpleReader simpleReader;
         inputSource.setData (strReply);
-        GvXMLParser xmlHandler;
+        HtmlFieldParser xmlHandler;
         xmlHandler.setEmitLog (emitLog);
 
         simpleReader.setContentHandler (&xmlHandler);
@@ -1400,8 +1512,16 @@ GVApi::onGetInbox(bool success, const QByteArray &response, QNetworkReply *,
             break;
         }
 
+        if (!xmlHandler.elems.contains ("json") ||
+            !xmlHandler.elems.contains ("html")) {
+            Q_WARN("Couldn't parse either the JSON or the HTML");
+            break;
+        }
+
         qint32 msgCount = 0;
-        if (!parseInboxJson(token, xmlHandler.strJson, xmlHandler.strHtml,
+        if (!parseInboxJson(token,
+                            xmlHandler.elems["json"].toString(),
+                            xmlHandler.elems["html"].toString(),
                             msgCount)) {
             Q_WARN("Failed to parse GV Inbox JSON. Data =") << strReply;
             break;
@@ -2451,7 +2571,7 @@ GVApi::onCheckRecentInbox(bool success, const QByteArray &response,
         QXmlInputSource inputSource;
         QXmlSimpleReader simpleReader;
         inputSource.setData (strReply);
-        GvXMLParser xmlHandler;
+        HtmlFieldParser xmlHandler;
         xmlHandler.setEmitLog (emitLog);
 
         simpleReader.setContentHandler (&xmlHandler);
@@ -2462,13 +2582,20 @@ GVApi::onCheckRecentInbox(bool success, const QByteArray &response,
             break;
         }
 
-        QString strTemp;
-        strTemp = "var obj = " + xmlHandler.strJson;
+        if (!xmlHandler.elems.contains ("json") ||
+            !xmlHandler.elems.contains ("html")) {
+            Q_WARN("Couldn't parse either the JSON or the HTML");
+            break;
+        }
+
+        QString strTemp, strJson;
+        strJson = xmlHandler.elems["json"].toString();
+        strTemp = "var obj = " + strJson;
         scriptEngine.evaluate (strTemp);
         if (scriptEngine.hasUncaughtException ()) {
             Q_WARN("Failed to assign json to obj. error =")
                << scriptEngine.uncaughtException().toString ()
-               << "JSON =" << xmlHandler.strJson;
+               << "JSON =" << strJson;
             break;
         }
 
@@ -2481,7 +2608,7 @@ GVApi::onCheckRecentInbox(bool success, const QByteArray &response,
         if (scriptEngine.hasUncaughtException ()) {
             Q_WARN("Uncaught exception executing script :")
                 << scriptEngine.uncaughtException().toString()
-                << "JSON =" << xmlHandler.strJson;
+                << "JSON =" << strJson;
             break;
         }
 
@@ -2493,7 +2620,7 @@ GVApi::onCheckRecentInbox(bool success, const QByteArray &response,
             if (scriptEngine.hasUncaughtException ()) {
                 Q_WARN("Uncaught exception executing script :")
                     << scriptEngine.uncaughtException().toString()
-                    << "JSON =" << xmlHandler.strJson;
+                    << "JSON =" << strJson;
                 break;
             }
 
@@ -2514,7 +2641,7 @@ GVApi::onCheckRecentInbox(bool success, const QByteArray &response,
         if (scriptEngine.hasUncaughtException ()) {
             Q_WARN("Uncaught exception executing script :")
                 << scriptEngine.uncaughtException().toString()
-                << "JSON =" << xmlHandler.strJson;
+                << "JSON =" << strJson;
             break;
         }
 
