@@ -21,7 +21,7 @@ Contact: yuvraaj@gmail.com
 
 #include "OsDependant.h"
 
-#if DESKTOP_OS
+#if DESKTOP_OS && SKYPE_ENABLED
 #include "SkypeClientFactory.h"
 #endif
 
@@ -33,16 +33,24 @@ createOSD(QObject *parent /* = NULL*/)
 
 OsDependant::OsDependant(QObject *parent /* = NULL*/)
 : IOsDependant(parent)
-#if DESKTOP_OS
+#if DESKTOP_OS && SKYPE_ENABLED
 , m_skypeClientFactory(new SkypeClientFactory(this))
 #endif
 {
 }//OsDependant::OsDependant
 
-#if DESKTOP_OS
+#if DESKTOP_OS && SKYPE_ENABLED
 SkypeClientFactory &
 OsDependant::skypeClientFactory()
 {
     return (*m_skypeClientFactory);
 }//OsDependant::skypeClientFactory
 #endif
+
+void
+OsDependant::setMainWidget (QWidget *win)
+{
+#if DESKTOP_OS && SKYPE_ENABLED
+    m_skypeClientFactory->setMainWidget (win);
+#endif
+}//OsDependant::setMainWidget
